/// @file   include/dmlite/cpp/dmlite.h
/// @brief  Entry point for DMLite.
/// @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
#ifndef DMLITE_CPP_DMLITE_H
#define DMLITE_CPP_DMLITE_H

#include "dmlite/common/config.h"
#include "exceptions.h"

#include <boost/any.hpp>
#include <list>
#include <map>
#include <string>
#include "utils/logger.h"

/// Namespace for the dmlite C++ API
namespace dmlite {

  /// API Version.
  const unsigned API_VERSION = 20121218;

  // Forward declarations.
  class Authn;
  class AuthnFactory;
  class BaseFactory;
  class Catalog;
  class CatalogFactory;
  class INode;
  class INodeFactory;
  class IODriver;
  class IODriverFactory;
  class PoolDriver;
  class PoolDriverFactory;
  class PoolManager;
  class PoolManagerFactory;
  class SecurityContext;  
  class SecurityCredentials;
  
  class StackInstance;

  /// CatalogInterface can only be instantiated through this class.
  class PluginManager {
   public:
    /// Constructor
    PluginManager() throw ();

    /// Destructor
    ~PluginManager();

    /// Load a plugin. Previously instantiated interfaces won't be affected.
    /// @param lib     The .so file. Usually, (path)/plugin_name.so.
    /// @param id      The plugin ID. Usually, plugin_name.
    void loadPlugin(const std::string& lib,
                    const std::string& id)  ;

    /// Set a configuration parameter. It will be passed to the loaded plugins.
    /// @param key   The configuration parameter.
    /// @param value The value for the configuration parameter.
    void configure(const std::string& key,
                   const std::string& value)  ;

    /// Load a configuration file, with plugins and parameters.
    /// @param file The configuration file.
    void loadConfiguration(const std::string& file)  ;

    /// Return an entry from the loaded configuration.
    /// @param key The configuration parameter.
    std::string getConfiguration(const std::string& key)  ;

    /// Register a Authn factory. To be used by concrete implementations
    /// @param factory The UserDbGroup concrete factory.
    /// @note  The same object can be passed to other register functions.
    ///        DMLite will take care of freeing it only once.
    void registerAuthnFactory(AuthnFactory* factory)  ;

    /// Register a INode factory. To be used by concrete implementations (i.e. Plugins)
    /// @param factory The INode concrete factory.
    /// @note  The same object can be passed to other register functions.
    ///        DMLite will take care of freeing it only once.
    void registerINodeFactory(INodeFactory* factory)  ;

    /// Register a catalog factory. To be used by concrete implementations (i.e. Plugins)
    /// @param factory The catalog concrete factory.
    /// @note  The same object can be passed to other register functions.
    ///        DMLite will take care of freeing it only once.
    void registerCatalogFactory(CatalogFactory* factory)  ;

    /// Register a pool factory.
    /// @param factory The pool concrete factory.
    /// @note  The same object can be passed to other register functions.
    ///        DMLite will take care of freeing it only once.
    void registerPoolManagerFactory(PoolManagerFactory* factory)  ;

    /// Register a IODriver factory.
    /// @param factory The IO concrete factory.
    /// @note  The same object can be passed to other register functions.
    ///        DMLite will take care of freeing it only once.
    void registerIODriverFactory(IODriverFactory* factory)  ;

    /// Register a PoolDriver factory.
    /// @param factory The PoolDriver factory.
    /// @note  The same object can be passed to other register functions.
    ///        DMLite will take care of freeing it only once.
    void registerPoolDriverFactory(PoolDriverFactory* factory)  ;
    
    /// Register a bare BaseFactory. Only the configure method will be called.
    /// @param factory The BaseFactory.
    /// @note  The same object can be passed to other register functions.
    ///        DMLite will take care of freeing it only once.
    void registerConfigureFactory(BaseFactory* factory)  ;
    
    /// Get the AuthnFactory implementation on top of the plugin stack.
    AuthnFactory* getAuthnFactory()  ;

    // Get the INodeFactory implementation on top of the plugin stack.
    INodeFactory* getINodeFactory()  ;

    /// Get the CatalogFactory implementation on top of the plugin stack.
    CatalogFactory* getCatalogFactory()  ;

    /// Get the PoolFactory implementation on top of the plugin stack.
    PoolManagerFactory* getPoolManagerFactory()  ;

    /// Get the appropiate pool driver factory for the pool.
    PoolDriverFactory* getPoolDriverFactory(const std::string& pooltype)  ;

    /// Get the IOFactory implementation on top of the plugin stack.
    IODriverFactory* getIODriverFactory()  ;

   private:
    /// Configuration key/value
    std::map<std::string, std::string> confValues_;

    /// Internal list of loaded plug-ins.
    std::list<AuthnFactory*>       authn_plugins_;
    std::list<INodeFactory*>       inode_plugins_;
    std::list<CatalogFactory*>     catalog_plugins_;
    std::list<PoolManagerFactory*> pool_plugins_;
    std::list<IODriverFactory*>    io_plugins_;
    std::list<PoolDriverFactory*>  pool_driver_plugins_;
    std::list<BaseFactory*>        configure_factory_;

    /// Keep pointers returned by dlopen at hand to free on destruction
    std::list<void*> dlHandles_;

    /// Can not be copied
    PluginManager(const PluginManager&);
  };

  extern Logger::bitmask stackinstancelogmask;
  extern Logger::component stackinstancelogname;
  
  /// We need to have something that allows one plugin stack to access
  /// another plugin stack, so this represents a instantiation
  /// of each plugin stack.
  /// It also keeps common state: user credentials, security context,
  /// and run-time parameters (see set)
  /// @note Assume a StackInstance (and every instantiated interface under it)
  ///       is NOT thread-safe. This means, a StackInstance must be used by only
  ///       one thread at the same time.
  class StackInstance {
   public:
    /// Constructor.
    StackInstance(PluginManager* pm)  ;

    /// Destructor.
    ~StackInstance();

    /// Set a key-value pair associated with this context.
    /// This can be used to pass advanced parameters to and from the plugins.
    /// @param key   The key.
    /// @param value The value.
    void set(const std::string& key, const boost::any& value)  ;

    /// Get a value associated to a key.
    /// This can be used to pass advanced parameters to and from the plugins.
    /// @param key The key parameter.
    boost::any get(const std::string& key) const  ;

    /// Erase a key,value pair from.
    /// @param key The key of the pair to be erased.
    void erase(const std::string& key)  ;

    /// Erase all the values set previously.
    void eraseAll(void) throw ();

    /// Checks if the stack instance contains a value associated with
    /// the given key.
    bool contains(const std::string& key) throw ();

    /// Get the plugin manager.
    PluginManager* getPluginManager()  ;

    /// Set the security credentials.
    void setSecurityCredentials(const SecurityCredentials& cred)  ;

    /// Set the security context.
    void setSecurityContext(const SecurityContext& ctx)  ;

    /// Return the security context.
    const SecurityContext* getSecurityContext(void) const throw ();

    /// Get the UsersDb interface.
    Authn* getAuthn()  ;

    /// Get the INode.
    INode* getINode()  ;

    /// Get the catalog.
    Catalog* getCatalog()  ;

    // Check if there is a PoolManager available
    bool isTherePoolManager() throw ();

    /// Get the PoolManager.
    PoolManager* getPoolManager()  ;

    /// Get a pool driver.
    PoolDriver* getPoolDriver(const std::string& poolType)  ;

    /// Get the IO driver.
    IODriver* getIODriver()  ;

  private:
    PluginManager* pluginManager_;

    Authn*       authn_;
    INode*       inode_;
    Catalog*     catalog_;
    PoolManager* poolManager_;
    IODriver*    ioDriver_;

    SecurityContext* secCtx_;

    std::map<std::string, PoolDriver*> poolDrivers_;

    std::map<std::string, boost::any> stackMsg_;

    void setSecurityContextImpl_(void);
  };

  /// Joint between plugins and plugin-manager
  struct PluginIdCard {
    /// Used to make sure API is consistent.
    unsigned ApiVersion;
    /// Let the plug-in register itself and its concrete factories
    void (*registerPlugin)(PluginManager* pm)  ;
    
  };

  /// Macro intended to allow future expansions of the PluginIdCard header
  /// easily.
  #define PLUGIN_ID_HEADER dmlite::API_VERSION

};

#endif // DMLITE_CPP_DMLITE_H
