/*
 * Copyright 2015 CERN
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */



/** @file   DomeCore.cpp
 * @brief  The main class where operation requests are applied to the status
 * @author Fabrizio Furano
 * @date   Dec 2015
 */


#include "DomeCore.h"
#include "DomeLog.h"
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <sys/vfs.h>
#include <unistd.h>
//#include <fastcgi.h>
#include <openssl/crypto.h>

#include <sys/time.h>
#include <sys/resource.h>

DomeCore::DomeCore() {
  domelogmask = Logger::get()->getMask(domelogname);
  //Info(Logger::Lvl1, fname, "Ctor " << dmlite_MAJOR <<"." << dmlite_MINOR << "." << dmlite_PATCH);
  initdone = false;
  terminationrequested = false;

}

DomeCore::~DomeCore() {


  Log(Logger::Lvl1, domelogmask, domelogname, "Stopping ticker.");


  if(davixPool) {
    delete davixPool;
    davixPool = NULL;
  }

  if(davixFactory) {
    delete davixFactory;
    davixFactory = NULL;
  }

  if (ticker) {
    Log(Logger::Lvl1, domelogmask, domelogname, "Joining ticker.");
    ticker->interrupt();
    ticker->join();
    delete ticker;
    ticker = 0;
    Log(Logger::Lvl1, domelogmask, domelogname, "Joined ticker.");

  }


}





int DomeCore::processreq(DomeReq &dreq) {
    
    status.stats.countRequest();
    
    // Last barrier against uncatched DmExceptions
    // If any comes, we give a generic error citing it
    try {
      try {
        try {
          
  
          Log(Logger::Lvl4, domelogmask, domelogname, "clientdn: '" << dreq.clientdn << "' clienthost: '" << dreq.clienthost <<
          "' remoteclient: '" << dreq.creds.clientName << "' remoteclienthost: '" << dreq.creds.remoteAddress);
          
          Log(Logger::Lvl4, domelogmask, domelogname, "req:" << dreq.verb << " cmd:" << dreq.domecmd << " query:" << dreq.object << " bodyitems: " << dreq.bodyfields.size() );
          
          
          // -------------------------
          // Generic authorization
          // Please note that authentication must be configured in the web server, not in DOME
          // -------------------------
          
          // First, check the dn is whitelisted as 'root' user
          bool authorize = status.isDNwhitelisted((char *)dreq.clientdn.c_str());

          
          if (!authorize) {
            // The whitelist in the config file did not authorize
            // Anyway this call may come from a server that was implicitly known, e.g.
            // head node trusts all the disk nodes that are registered in the filesystem table
            // disk node trusts head node as defined in the config file
            
            {
              boost::unique_lock<boost::recursive_mutex> l(status);
              authorize = status.isDNaKnownServer(dreq.clientdn);
            }
            if (authorize)
              Log(Logger::Lvl2, domelogmask, domelogname, "DN '" << dreq.clientdn << "' is authorized as a known server of this cluster.");
          }
          
          // -------------------------
          // Command dispatching
          // -------------------------
          
          if (authorize) {
            
            // Client was authorized. We log the request
            Log(Logger::Lvl1, domelogmask, domelogname, "clientdn: '" << dreq.clientdn << "' clienthost: '" << dreq.clienthost <<
            "' remoteclient: '" << dreq.creds.clientName << "' remoteclienthost: '" << dreq.creds.remoteAddress << "'");
            
            
            
            if (dreq.bodyfields.get<std::string>("lfn", "").size() > 0)
              Log(Logger::Lvl1, domelogmask, domelogname, "req:" << dreq.verb << " cmd:" << dreq.domecmd <<
              " query:" << dreq.object << " bodyitems: " << dreq.bodyfields.size() << " lfn: '" << dreq.bodyfields.get<std::string>("lfn", "") << "'");
            else
              Log(Logger::Lvl1, domelogmask, domelogname, "req:" << dreq.verb << " cmd:" << dreq.domecmd << " query:" << dreq.object << " bodyitems: " << dreq.bodyfields.size());
            
            // First discriminate on the HTTP request: GET/POST, etc..
            if(dreq.verb == "GET") {
              
              // Now dispatch based on the actual command name
              if ( dreq.domecmd == "dome_access" ) {
                dome_access(dreq);
              } else if ( dreq.domecmd == "dome_statpfn" ) {
                dome_statpfn(dreq);
              } else if ( dreq.domecmd == "dome_getstatinfo" ) {
                dome_getstatinfo(dreq);
              } else if ( dreq.domecmd == "dome_getreplicainfo" ) {
                dome_getreplicainfo(dreq);
              } else if ( dreq.domecmd == "dome_accessreplica" ) {
                dome_accessreplica(dreq);
              } else if ( dreq.domecmd == "dome_getspaceinfo" ) {
                dome_getspaceinfo(dreq);
              } else if(dreq.domecmd == "dome_chksum") {
                dome_chksum(dreq);
              } else if(dreq.domecmd == "dome_getdirspaces") {
                dome_getdirspaces(dreq);
              }else if(dreq.domecmd == "dome_getquotatoken") {
                dome_getquotatoken(dreq);
              }else if(dreq.domecmd == "dome_get") {
                dome_get(dreq);
              } else if ( dreq.domecmd == "dome_statpool" ) {
                dome_statpool(dreq);
              } else if ( dreq.domecmd == "dome_getdir" ) {
                dome_getdir(dreq);
              } else if ( dreq.domecmd == "dome_getuser" ) {
                dome_getuser(dreq);
              } else if ( dreq.domecmd == "dome_getusersvec" ) {
                dome_getusersvec(dreq);
              } else if ( dreq.domecmd == "dome_getidmap" ) {
                dome_getidmap(dreq);
              } else if (dreq.domecmd == "dome_info") {
                dome_info(dreq, -1, authorize);
              } else if (dreq.domecmd == "dome_getcomment") {
                dome_getcomment(dreq);
              } else if (dreq.domecmd == "dome_getgroup") {
                dome_getgroup(dreq);
              } else if (dreq.domecmd == "dome_getgroupsvec") {
                dome_getgroupsvec(dreq);
              } else if (dreq.domecmd == "dome_getreplicavec") {
                dome_getreplicavec(dreq);
              } else if (dreq.domecmd == "dome_readlink") {
                dome_readlink(dreq);
              } else {
                dreq.SendSimpleResp(418, SSTR("Command '" << dreq.object << "' unknown for a GET request. I like your style."));
                return -1;
              }
              
            } else if(dreq.verb == "HEAD"){ // meaningless placeholder
              dreq.SendSimpleResp(200, SSTR("This is contradicting, isn't it ?"));
              return -1;
              
            } else if(dreq.verb == "POST"){
              if ( dreq.domecmd == "dome_put" ) {
                bool success;
                dome_put(dreq, success);
              }
              else if ( dreq.domecmd == "dome_putdone" ) {
                
                if(status.role == status.roleHead)
                  dome_putdone_head(dreq);
                else
                  dome_putdone_disk(dreq);
                
              }
              else if (dreq.domecmd == "dome_unlink") {
                dome_unlink(dreq);
              }
              else if ( dreq.domecmd == "dome_setquotatoken" ) {
                dome_setquotatoken(dreq);
              }
              else if ( dreq.domecmd == "dome_addreplica" ) {
                dome_addreplica(dreq);
              }
              else if ( dreq.domecmd == "dome_delreplica" ) {
                dome_delreplica(dreq);
              }
              else if ( dreq.domecmd == "dome_pfnrm" ) {
                dome_pfnrm(dreq);
              }
              else if ( dreq.domecmd == "dome_addfstopool" ) {
                dome_addfstopool(dreq);
              }
              else if ( dreq.domecmd == "dome_modifyfs" ) {
                dome_modifyfs(dreq);
              }
              else if ( dreq.domecmd == "dome_rmfs" ) {
                dome_rmfs(dreq);
              }
              else if ( dreq.domecmd == "dome_delquotatoken" ) {
                dome_delquotatoken(dreq);
              }
              else if(dreq.domecmd == "dome_chksumstatus") {
                dome_chksumstatus(dreq);
              }
              else if(dreq.domecmd == "dome_dochksum") {
                dome_dochksum(dreq);
              }
              else if(dreq.domecmd == "dome_addfstopool") {
                dome_addfstopool(dreq);
              }
              else if(dreq.domecmd == "dome_rmpool") {
                dome_rmpool(dreq);
              }
              else if(dreq.domecmd == "dome_addpool") {
                dome_addpool(dreq);
              }
              else if(dreq.domecmd == "dome_modifypool") {
                dome_modifypool(dreq);
              }
              else if(dreq.domecmd == "dome_pull") {
                dome_pull(dreq);
              }
              else if(dreq.domecmd == "dome_pullstatus") {
                dome_pullstatus(dreq);
              }
              else if(dreq.domecmd == "dome_updatexattr") {
                dome_updatexattr(dreq);
              }
              else if(dreq.domecmd == "dome_makespace") {
                dome_makespace(dreq);
              }
              else if(dreq.domecmd == "dome_modquotatoken") {
                dome_modquotatoken(dreq);
              }
              else if(dreq.domecmd == "dome_create") {
                dome_create(dreq);
              }
              else if(dreq.domecmd == "dome_makedir") {
                dome_makedir(dreq);
              }
              else if(dreq.domecmd == "dome_deleteuser") {
                dome_deleteuser(dreq);
              }
              else if(dreq.domecmd == "dome_newuser") {
                dome_newuser(dreq);
              }
              else if(dreq.domecmd == "dome_updateuser") {
                dome_updateuser(dreq);
              }
              else if(dreq.domecmd == "dome_deletegroup") {
                dome_deletegroup(dreq);
              }
              else if(dreq.domecmd == "dome_newgroup") {
                dome_newgroup(dreq);
              }
              else if(dreq.domecmd == "dome_updategroup") {
                dome_updategroup(dreq);
              }
              else if(dreq.domecmd == "dome_setcomment") {
                dome_setcomment(dreq);
              }
              else if (dreq.domecmd == "dome_removedir") {
                dome_removedir(dreq);
              }
              else if (dreq.domecmd == "dome_symlink") {
                dome_symlink(dreq);
              }
              else if (dreq.domecmd == "dome_rename") {
                dome_rename(dreq);
              }
              else if (dreq.domecmd == "dome_setacl") {
                dome_setacl(dreq);
              }
              else if (dreq.domecmd == "dome_setmode") {
                dome_setmode(dreq);
              }
              else if (dreq.domecmd == "dome_setowner") {
                dome_setowner(dreq);
              }
              else if (dreq.domecmd == "dome_setsize") {
                dome_setsize(dreq);
              }
              else if (dreq.domecmd == "dome_setchecksum") {
                dome_setchecksum(dreq);
              }
              else if (dreq.domecmd == "dome_updatereplica") {
                dome_updatereplica(dreq);
              }
              else {
                dreq.SendSimpleResp(418, SSTR("Command '" << dreq.domecmd << "' unknown for a POST request.  Nice joke, eh ?"));
                return -1;
              }
            }
            
          } // if authorized
          else {
            // only possible to run info when unauthorized
            if(dreq.domecmd == "dome_info") {
              dome_info(dreq, -1, authorize);
            }
            else {
              Err(domelogname, "DN '" << dreq.clientdn << "' has NOT been authorized.");
              dreq.SendSimpleResp(403, SSTR(dreq.clientdn << " is unauthorized. Sorry :-)"));
              return -1;
            }
          }
          
        } catch (dmlite::DmException e) {
          Err(domelogname, "Wrong parameters. err: " << e.code() << " what: '" << e.what() << "'");
          dreq.SendSimpleResp(422, SSTR("Wrong parameters. err: " << e.code() << " what: '" << e.what() << "'"));
          return -1;
        }
      }
      catch(boost::property_tree::ptree_error &e) {
        Err(domelogname, "Error while parsing json body: " << e.what());
        dreq.SendSimpleResp(422, SSTR("Error while parsing json body: " << e.what()));
        return -1;
      }
    }
    catch(...) {
      Err(domelogname, "Generic exception.");
      dreq.SendSimpleResp(422, "Generic exception.");
      return -1;
    }
    
  
  return 0;
}








static Davix::RequestParams getDavixParams() {
  Davix::RequestParams params;

  // set timeouts, etc
  long conn_timeout = CFG->GetLong("glb.restclient.conn_timeout", 15);
  struct timespec spec_timeout;
  spec_timeout.tv_sec = conn_timeout;
  spec_timeout.tv_nsec = 0;
  params.setConnectionTimeout(&spec_timeout);
  Log(Logger::Lvl1, domelogmask, domelogname, "Davix: Connection timeout is set to : " << conn_timeout);

  long ops_timeout = CFG->GetLong("glb.restclient.ops_timeout", 15);
  spec_timeout.tv_sec = ops_timeout;
  spec_timeout.tv_nsec = 0;
  params.setOperationTimeout(&spec_timeout);
  Log(Logger::Lvl1, domelogmask, domelogname, "Davix: Operation timeout is set to : " << ops_timeout);

  // get ssl check
  bool ssl_check = CFG->GetBool("glb.restclient.ssl_check", true);
  params.setSSLCAcheck(ssl_check);
  Log(Logger::Lvl1, domelogmask, domelogname, "SSL CA check for davix is set to  " + std::string((ssl_check) ? "TRUE" : "FALSE"));

  // ca check
  std::string ca_path = CFG->GetString("glb.restclient.ca_path", (char *)"/etc/grid-security/certificates");
  if( !ca_path.empty()) {
    Log(Logger::Lvl1, domelogmask, domelogname, "CA Path :  " << ca_path);
    params.addCertificateAuthorityPath(ca_path);
  }

  // credentials
  Davix::X509Credential cred;
  Davix::DavixError* tmp_err = NULL;

  
  cred.loadFromFilePEM(CFG->GetString("glb.restclient.cli_private_key", (char *)""), CFG->GetString("glb.restclient.cli_certificate", (char *)""), "", &tmp_err);
  if( tmp_err ){
      std::ostringstream os;
      os << "Cannot load cert-privkey " << CFG->GetString("glb.restclient.cli_certificate", (char *)"") << "-" <<
            CFG->GetString("glbDomeCore::.restclient.cli_private_key", (char *)"") << ", Error: "<< tmp_err->getErrMsg();

      Davix::DavixError::clearError(&tmp_err);
      throw dmlite::DmException(EPERM, os.str());
  }
  params.setClientCertX509(cred);
  
  
  params.setKeepAlive(CFG->GetBool("glb.restclient.keepalive", true));
  
  
  return params;
}

int DomeCore::init(const char *cfgfile) {
  const char *fname = "UgrConnector::init";
  {
    boost::lock_guard<boost::recursive_mutex> l(mtx);
    if (initdone) return -1;

    Logger::get()->setLevel(Logger::Lvl4);

    // Process the config file
    Log(Logger::Lvl1, domelogmask, domelogname, "------------------------------------------");
    Log(Logger::Lvl1, domelogmask, domelogname, "------------ Starting. Config: " << cfgfile);

    struct rlimit rlim;
    if (getrlimit(RLIMIT_NOFILE, &rlim)) {
      Err(fname, "Initialization error. Cannot invoke getrlimit()");
      return -1;
    }
    Log(Logger::Lvl1, domelogmask, domelogname, "------------ File descriptors: " << rlim.rlim_cur <<
      " max: " << rlim.rlim_max);
    
    if (rlim.rlim_cur < 48*1024) {
      Err(fname, "Initialization error. Not enough file descriptors. Please apply the system tuning hints.");
      return -1;
    }
    
    if (!cfgfile || !strlen(cfgfile)) {
      Err(fname, "No config file given." << cfgfile);
      return -1;
    }

    if (CFG->ProcessFile((char *)cfgfile)) {
      Err(fname, "Error processing config file." << cfgfile << std::endl);
      return 1;
    }

    // Initialize the logger
    long debuglevel = CFG->GetLong("glb.debug", 1);
    Logger::get()->setLevel((Logger::Level)debuglevel);

    // Initialize the metadata cache
    DomeMetadataCache *dmc = DOMECACHE;
    if (dmc) {
      dmc->Init();
      Log(Logger::Lvl1, domelogmask, domelogname, "Cache successfully started. maxitems: " <<
      CFG->GetLong("mdcache.maxitems", 100000) << " itemttl:" << CFG->GetLong("mdcache.itemttl", 3600));
    }
    else
      Log(Logger::Lvl1, domelogmask, domelogname, "Could not start the DOME cache.");
    
    std::string r = CFG->GetString("glb.role", (char *)"head");
    if (r == "head") status.role = status.roleHead;
    else if (r == "disk") status.role = status.roleDisk;
    else {
      Err(fname, "Invalid role: '" << r << "'");
      return -1;
    }

    if (status.role == status.roleHead)
      status.headnodename = status.myhostname;
    else {
      // Now get the host name of the head node
      Davix::Uri uri(CFG->GetString("disk.headnode.domeurl", (char *)""));
      status.headnodename = uri.getHost();
    }

    Log(Logger::Lvl1, domelogmask, domelogname, "My head node hostname is: '" << status.headnodename << "'");

    // The limits for the prio queues, get them from the cfg
    std::vector<size_t> limits;
    limits.push_back( CFG->GetLong("head.checksum.maxtotal", 1000) );
    limits.push_back( CFG->GetLong("head.checksum.maxpernode", 40) );

    // Create the chksum queue
    status.checksumq = new GenPrioQueue(CFG->GetLong("head.checksum.qtmout", 15), limits);

    // Create the queue for the callouts
    limits.clear();
    limits.push_back( CFG->GetLong("head.filepulls.maxtotal", 1000) );
    limits.push_back( CFG->GetLong("head.filepulls.maxpernode", 40) );
    status.filepullq = new GenPrioQueue(CFG->GetLong("head.filepulls.qtmout", 15), limits);

    // Allocate the mysql factory and configure it
    if(status.role == status.roleHead) {
      DomeMySql::configure( CFG->GetString("head.db.host",     (char *)"localhost"),
                            CFG->GetString("head.db.user",     (char *)"guest"),
                            CFG->GetString("head.db.password", (char *)"none"),
                            CFG->GetLong  ("head.db.port",     0),
                            CFG->GetLong  ("head.db.poolsz",   128),
                            CFG->GetString("head.db.cnsdbname", (char *)"cns_db"),
                            CFG->GetString("head.db.dpmdbname", (char *)"dpm_db")
                          );

      // Try getting a db connection and use it. If it does not work
      // an exception will just kill us, which is what we want
      DomeMySql sql;
      // Check for the presence of the root entry in the db. If error... exit
      dmlite::DmStatus ret = sql.checkRootEntry();
      if (!ret.ok()) {
        // If we can't create the dir then this is a serious error, unless it already exists
        Err(domelogname, "Cannot create root entry. err: " << ret.code() << "-" << ret.what());
        return -1;
      }

      // Only load quotatokens on the head node
      status.loadQuotatokens();
    }


    // Configure the davix pool
    davixFactory = new dmlite::DavixCtxFactory();

    davixFactory->setRequestParams(getDavixParams());
    davixPool = new dmlite::DavixCtxPool(davixFactory, CFG->GetLong("glb.restclient.poolsize", 256));
    status.setDavixPool(davixPool);

    // Load filesystems
    status.loadFilesystems();




    // Start the ticker
    Log(Logger::Lvl1, domelogmask, domelogname, "Starting ticker.");
    ticker = new boost::thread(boost::bind(&DomeCore::tick, this, 0));
    queueTicker = new boost::thread(boost::bind(&DomeCore::queueTick, this, 0));

    return 0;
  }
}



void DomeCore::tick(int parm) {

  while (! this->terminationrequested ) {
    time_t timenow = time(0);

    Log(Logger::Lvl4, domelogmask, domelogname, "Tick");

    try {
      status.tick(timenow);
      dmTaskExec::tick();
    
      DOMECACHE->tick();
      
      //
      // Print simple stats
      //
      float stats_reqrate = 0, db_reqrate = 0, db_transrate = 0, stats_intercluster = 0, peak_reqrate = 0;
      bool stats_print = false;
      time_t tdiff = time(0) - status.stats.lastcheck;
      time_t peaktdiff = time(0) - status.stats.lastpeakcheck;
      {
        boost::unique_lock<boost::mutex> l(status.stats);
        
        status.stats.peakreqrate = fmax(status.stats.peakreqrate, (float)status.stats.peakcntrequests / peaktdiff);
        status.stats.peakcntrequests = 0;
        status.stats.lastpeakcheck = time(0);
        
        if (tdiff > CFG->GetLong("glb.printstats.interval", 60)) {
          stats_reqrate = (float)status.stats.cntrequests / tdiff;
          stats_intercluster = (float)status.stats.intercluster / tdiff;
          peak_reqrate = status.stats.peakreqrate;
          stats_print = true;
          status.stats.lastcheck = time(0);
          status.stats.cntrequests = 0;
          status.stats.peakcntrequests = 0;
          status.stats.intercluster = 0;
          status.stats.peakreqrate = 0;
        }
      }
        
      
      if (stats_print) {
        {
          boost::unique_lock<boost::mutex> l(DomeMySql::dbstats);
          
          db_reqrate = (float)DomeMySql::dbstats.dbqueries / tdiff;
          db_transrate = (float)DomeMySql::dbstats.dbtrans / tdiff;
          DomeMySql::dbstats.dbqueries = 0;
          DomeMySql::dbstats.dbtrans = 0;
        }
        
        Log(Logger::Lvl1, domelogmask, domelogname, "Request rate: " << stats_reqrate << "Hz (Peak: " << peak_reqrate << "Hz) -- DB queries: " << db_reqrate << "Hz -- DB transactions: " << db_transrate   << "Hz -- Intercluster messages: " << stats_intercluster << "Hz");
      }
      
      
    }
    catch (...) {
      Err(domelogname, "Exception catched in the ticker. This is strange.");
    
    }

    sleep(CFG->GetLong("glb.tickfreq", 5));
  }

}

void DomeCore::queueTick(int parm) {

  while(! this->terminationrequested) {
    try {
      time_t timenow = time(0);
      status.waitQueues();

      Log(Logger::Lvl4, domelogmask, domelogname, "Tick");
      status.tickQueues(timenow);
    }
    catch (...) {
      Err(domelogname, "Exception catched in the queue ticker. This is strange.");
        
    }
  }
}

/// Send a notification to the head node about the completion of this task
void DomeCore::onTaskCompleted(dmlite::dmTask &task) {
  
  Log(Logger::Lvl4, domelogmask, domelogname, "Entering. key: " << task.key << "' rc: " << task.resultcode << "cmd: '" << task.cmd << "' stdout: '" << task.stdout << "'");
  
  if (task.resultcode)
    Err(domelogname, "Entering. key: " << task.key << "' rc: " << task.resultcode << "cmd: '" << task.cmd << "' stdout: '" << task.stdout << "'");
  
  int key = task.key;
  bool dosend = false;
  
  PendingChecksum pending;
  PendingPull pendingpull;
  
  std::map<int, PendingChecksum>::iterator it;
  {
    boost::lock_guard<boost::recursive_mutex> l(mtx);
    it = diskPendingChecksums.find(key);
    
    
    
    if(it != diskPendingChecksums.end()) {
      Log(Logger::Lvl4, domelogmask, domelogname, "Found pending checksum. key: " << task.key);
      pending = it->second;
      diskPendingChecksums.erase(it);
      dosend = true;
    }
  }
  
  if (dosend) {
    sendChecksumStatus(pending, task, true);
    Log(Logger::Lvl4, domelogmask, domelogname, "Entering. key: " << task.key);
    return;
  }
  
  dosend = false;
  std::map<int, PendingPull>::iterator pit;
  {
    boost::lock_guard<boost::recursive_mutex> l(mtx);
    pit = diskPendingPulls.find(key);
  
  
    if(pit != diskPendingPulls.end()) {
      pendingpull = pit->second;
      Log(Logger::Lvl4, domelogmask, domelogname, "Found pending file pull. key: " << task.key);
      diskPendingPulls.erase(pit);
      dosend = true;
    }
  }
  
  if (dosend) {
    sendFilepullStatus(pendingpull, task, true);
    return;
  }
  
  
  // This may be a stat
  // This would be an internal error!!!!
  //Err(domelogname, "Cannot match task notification. key: " << task.key);
}

/// Send a notification to the head node about a task that is running
void DomeCore::onTaskRunning(dmlite::dmTask &task) {
  Log(Logger::Lvl4, domelogmask, domelogname, "Entering. key: " << task.key);
  int key = task.key;
  PendingChecksum pending;
  PendingPull pendingpull;
  bool dosend = false;
  
  std::map<int, PendingChecksum>::iterator it;
  {
    boost::lock_guard<boost::recursive_mutex> l(mtx);
    it = diskPendingChecksums.find(key);
    
    if(it != diskPendingChecksums.end()) {
      pending = it->second;
      Log(Logger::Lvl4, domelogmask, domelogname, "Found pending checksum. key: " << task.key);
      dosend = true;
    }
  }
  
  if (dosend) {
    sendChecksumStatus(pending, task, false);
    return;
  }
  
  dosend = false;
  std::map<int, PendingPull>::iterator pit;
  {
    boost::lock_guard<boost::recursive_mutex> l(mtx);
    pit = diskPendingPulls.find(key);
    
    if(pit != diskPendingPulls.end()) {
      pendingpull = pit->second;
      dosend = true;
      Log(Logger::Lvl4, domelogmask, domelogname, "Found pending file pull. key: " << task.key);
    }
  }
  
  if (dosend) {
    sendFilepullStatus(pendingpull, task, false);
    return;
  }
  
  // This would be an internal error or just a stat request.
  //Err(domelogname, "Cannot match task notification. key: " << task.key);
}


void DomeCore::fillSecurityContext(dmlite::SecurityContext &ctx, DomeReq &req) {

  // Take the info coming from the request
  req.fillSecurityContext(ctx);

  Log(Logger::Lvl4, domelogmask, domelogname,
      "clientdn: '" << ctx.credentials.clientName << "' " <<
      "clienthost: '" << ctx.credentials.remoteAddress << "' " <<
      "ctx.user.name: '" << ctx.user.name << "' " <<
      "ctx.groups: " << ctx.groups.size() << "(size) "
  );

  // Now map uid and gids into the spooky extensible
  if (status.isDNRoot(ctx.user.name.c_str())) {
    // A client can connect ONLY if it has the right identity, i.e.
    // it's a machine in the cluster
    // If on top of that such a client has a privileged DN
    // then he is root, and can read/write everywhere
    // A privileged DN is a DN that has been explicitely whitelisted
    // OR the local hostname
    ctx.user["uid"] = 0;
    ctx.user["gid"] = 0;
    ctx.user["banned"] = false;
  }
  else {
    DomeUserInfo u;
    if (status.getUser(ctx.user.name, u)) {
      ctx.user["uid"] = u.userid;
      ctx.user["banned"] = (int)u.banned;
    } else {
      // Maybe we have to do something if the user was unknown?
      DomeMySql sql;
      if ((ctx.user.name.length() > 0) && sql.newUser(u, ctx.user.name).ok()) {
        ctx.user["uid"] = u.userid;
        ctx.user["banned"] = (int)u.banned;
      }
      else
        Err(domelogname, "Cannot add unknown user '" << ctx.user.name << "'");
    }

    DomeGroupInfo g;
    for(size_t i = 0; i < ctx.groups.size(); i++) {
      if (status.getGroup(ctx.groups[i].name, g)) {
        ctx.groups[i]["gid"] = g.groupid;
        ctx.groups[i]["banned"] = (int)g.banned;
      } else {
        // Maybe we have to do something if the group was unknown?
        DomeMySql sql;
        if ((ctx.groups[i].name.length() > 0) && sql.newGroup(g, ctx.groups[i].name).ok()) {
          ctx.groups[i]["gid"] = g.groupid;
          ctx.groups[i]["banned"] = (int)g.banned;
        }
        else
          Err(domelogname, "Cannot add unknown group '" << ctx.groups[i].name << "'");
      }


    }
  }

}
