
/*
 * Copyright 2015 CERN
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 */



/** @file   dmTaskExec.cpp
 * @brief  A class that spawns commands that perform actions
 * @author Fabrizio Furano
 * @author Andrea Manzi
 * @date   Dec 2015
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include "utils/logger.h"
#include "utils/Config.hh"
#include "TaskExec.h"
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/thread.hpp>
#include <queue>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>

using namespace boost;
using namespace std;



#define dmTaskLog(inst, lvl, where, what)                      \
{                                                              \
  if (Logger::get()->getLevel() >= lvl) {                      \
    std::ostringstream outs;                                   \
    outs << where << " " << __func__ << " : " << what;         \
    inst->onLoggingRequest((Logger::Level)lvl, outs.str());    \
  }                                                            \
}                                                              \
\

#define dmTaskErr(inst, where, what)                           \
{                                                              \
  std::ostringstream outs;                                     \
  outs << where << " " << __func__ << " : " << what;           \
  inst->onErrLoggingRequest(outs.str());                       \
}                                                              \
\



namespace dmlite {
  
  void taskfunc(dmTaskExec *inst, int key) {
    
    dmTaskLog(inst, Logger::Lvl4, "taskfunc", "Starting task " << key << " on instance " << inst->instance);
    
    if (inst) {
      map <int, dmTask *>::iterator i = inst->tasks.find(key);
      if ( i != inst->tasks.end() ) {
        dmTaskLog(inst, Logger::Lvl3, "taskfunc", "Found task " << key << " on instance " << inst->instance);
        inst->run(*i->second);
        dmTaskLog(inst, Logger::Lvl3, "taskfunc", "Finished task " << key << " on instance " << inst->instance);
        return;
      }
      
    }  
    dmTaskErr(inst, "taskfunc", "Cannot start task " << key << " on instance " << inst->instance);
    
    
  }
  
  // -----------------------------------------------------------------------------------
  // -----------------------------------------------------------------------------------
  // -----------------------------------------------------------------------------------
  // -----------------------------------------------------------------------------------
  
  dmTask::dmTask(dmTaskExec *wheretolog): finished(false), loggerinst(wheretolog) {
    starttime = time(0);
    endtime = 0;
    pid = -1;
    
    // Initialize placeholders for individual parms for execv
    for (int i =0; i < 64; i++)
      parms[i] = NULL;
  }
  
  dmTask::~dmTask() {
    //boost::unique_lock <boost::mutex> lck (*this);
    for (int i =0; i < 64; i++)
      // Free the execv parms, if we need...
      if (parms[i]) free((void *)parms[i]);
      else break;
  }
  
  
  void dmTask::splitCmd()
  {
    const char *tok;
    char *saveptr, *str = (char *)cmd.c_str();
    int i = 0;
    
    while ( (tok = strtok_r(str, " ", &saveptr)) ) {
      parms[i++] = strdup(tok);
      str = 0;
    }
    
    return;
    
  }
  
  
  int dmTask::waitFinished(int sectmout) {
    const char *fname = "dmTaskExec::waitFinished";
    
    dmTaskLog(loggerinst, Logger::Lvl4, fname, "Checking task termination. Key: " << key << " cmd: " << cmd << " finished: " << finished);
    
    system_time const timelimit = get_system_time() + posix_time::seconds(sectmout);
    dmTaskLog(loggerinst, Logger::Lvl4, fname, "Time: " << get_system_time());
    dmTaskLog(loggerinst, Logger::Lvl4, fname, "TimeLimit: " << timelimit);
    
    { 
      boost::unique_lock <boost::mutex>  lck (*this);
      while(!finished && get_system_time() < timelimit) {
        dmTaskLog(loggerinst, Logger::Lvl4, fname, "Task not finished at time " << get_system_time() << ": '" << cmd << "'");
        
        boost::system_time const tm = boost::get_system_time() + boost::posix_time::seconds(1);
        condvar.timed_wait(lck, tm);
      }
    }
    // We are here either if timeout or something happened
    
    if (finished) {
      dmTaskLog(loggerinst, Logger::Lvl3, fname, "Finished task. Key: " << key << " res: " << resultcode << " cmd: " << cmd);
      return 0;
    }
    
    dmTaskLog(loggerinst, Logger::Lvl3, fname, "Still running task. Key: " << key << " cmd: " << cmd);
    
    return 1;
    
  }
  
  
  // -----------------------------------------------------------------------------------
  // -----------------------------------------------------------------------------------
  // -----------------------------------------------------------------------------------
  // -----------------------------------------------------------------------------------
  dmTaskExec::dmTaskExec() {
    taskcnt = 1;
  }
  
  dmTaskExec::~dmTaskExec() {}
  
  /// Taken from here and slightly modified.
  /// https://sites.google.com/site/williamedwardscoder/popen3
  int dmTaskExec::popen3(int fd[3], pid_t *pid,  const  char ** argv) {
    int i, e;
    int p[3][2];
    //
    // set all the FDs to invalid
    for(i=0; i<3; i++)
      p[i][0] = p[i][1] = -1;
    
    // create the pipes
    for(int i=0; i<3; i++) 
      if(pipe(p[i]))
        goto error;
      
    // and fork
    *pid = fork();
    
    if(-1 == *pid)
      goto error;
    
    // in the parent?
    if(*pid) {
      // parent
      fd[STDIN_FILENO] = p[STDIN_FILENO][1];
      close(p[STDIN_FILENO][0]);
      
      fd[STDOUT_FILENO] = p[STDOUT_FILENO][0];
      close(p[STDOUT_FILENO][1]);
      
      fd[STDERR_FILENO] = p[STDERR_FILENO][0];
      close(p[STDERR_FILENO][1]);
      
      // success
      return 0;
      
    } else {
      // child
      while ( (dup2(p[STDIN_FILENO][0],STDIN_FILENO) == -1) && (errno == EINTR) ) {};
      close(p[STDIN_FILENO][1]);
      
      while ( (dup2(p[STDOUT_FILENO][1],STDOUT_FILENO) == -1) && (errno == EINTR) ) {};
      close(p[STDOUT_FILENO][0]);
      
      while ( (dup2(p[STDERR_FILENO][1],STDERR_FILENO) == -1) && (errno == EINTR) ) {};
      close(p[STDERR_FILENO][0]);
      
      int execret = execv(argv[0],(char **)argv);
      char errbuf[1024];
      errbuf[0] = '\0';
      strerror_r(errno, errbuf, 1024);
      errbuf[1023] = '\0';
      // if we are here, then we failed to launch our program
      dmTaskErr(this, "popen3", "Cannot launch cmd: " << argv[0] << " res: " << execret << " errno: " << errno << " err: '" << errbuf << "'");
      fprintf(stderr," \"%s\"\n",argv[0]);
      _exit(EXIT_FAILURE);
    }
    
    
    error:
    
    // preserve original error
    e = errno;
    for(i=0; i<3; i++) {
      close(p[i][0]);
      close(p[i][1]);
    }
    errno = e;
    return -1;
  }
  
  
  void dmTaskExec::run(dmTask &task) {
    
    dmTaskLog(this, Logger::Lvl3, "taskrun", "Starting command: '" << task.cmd << "'") ;
    //start time
    {
      boost::unique_lock <boost::mutex>  l(task);
      time(&task.starttime);
    }
    
    task.resultcode = popen3((int *)task.fd, &task.pid, task.parms);
    
    /// It is then possible for the parent process to read the output of the child process from file descriptor
    char buffer[256];

    
    while (1) {
      ssize_t count = read(task.fd[STDOUT_FILENO], buffer, sizeof(buffer));
      if (count == -1) {
        if (errno == EINTR) {
          continue;
        } else {
          dmTaskErr(this, "popen3", "Cannot get output of cmd: '" << task.cmd << "'");
          break;
        }
        
      }
      
      if (count == 0) {
        dmTaskLog(this, Logger::Lvl4, "taskrun", "End Stdout") ;
        break;
      }
      else {
        boost::unique_lock <boost::mutex>   l(task);
        task.stdout.append(buffer, count);
        
        if (Logger::get()->getLevel() >= Logger::Lvl4) {
          size_t pos = 0, pp = 0;
          int i = 0;
          // Break the large string into individual lines for logging
          while (1) {
            pp = task.stdout.find_first_of('\n', pos);

            if (pp == string::npos) {
              dmTaskLog(this, Logger::Lvl4, "taskrun",
                        "Pid " << task.pid << " stdout(" << i << "): '" <<
                        task.stdout.substr(pos, pp) << "'" );
              break;
            }
            else {
              dmTaskLog(this, Logger::Lvl4, "taskrun",
                        "Pid " << task.pid << " stdout(" << i << "): '" <<
                        task.stdout.substr(pos, (pp-pos)) << "'" );
            }
              
            pos = pp+1;
            i++;
          }
        }
      }
      
    }
    
    dmTaskLog(this, Logger::Lvl4, "taskrun", "Closing fds. key: " << task.key);
    for(int i=0; i<3; i++) 
      close(task.fd[i]);
    
    dmTaskLog(this, Logger::Lvl4, "taskrun", "Finalizing key: " << task.key) ;
    {
      boost::unique_lock <boost::mutex> l(task);
      int rescode;
      waitpid(task.pid, &rescode, 0);
      task.finished = true;
      
      if WIFSIGNALED(rescode) {
        dmTaskErr(this, "taskrun", "Task killed by signal " << WTERMSIG(rescode) << " command: '" << task.cmd << "'" );
        task.resultcode = -1;
      }
      else
        task.resultcode = WEXITSTATUS(rescode);
      
      //endtime
      time(&task.endtime);
      task.notifyAll();
      
      dmTaskLog(this, Logger::Lvl4, "taskrun", "Dispatching onTaskCompleted key: " << task.key << " command: '" << task.cmd << "'" );
      onTaskCompleted(task);
      
    }
    
    dmTaskLog(this, Logger::Lvl4, "taskrun", "Dispatched key: " << task.key) ;
    
    
    
  }
  
  
  int dmTaskExec::submitCmd(std::string cmd) {
    int taskkey;
    dmTask * task = new dmTask(this);
    task->cmd= cmd;
    task->splitCmd();
    {  
      scoped_lock lck (*this);
      
      task->key = ++taskcnt; 
      tasks.insert( std::pair<int,dmTask*>(task->key,task));
      taskkey = task->key;
    }
    boost::thread workerThread( taskfunc, this, taskcnt );
    return taskkey;
  }
  
  int dmTaskExec::submitCmd(std::vector<std::string> &args) {
    dmTask * task = NULL;
    int taskkey;
    std::ostringstream oss;
    
    if (!args.empty())
    {
      task =  new dmTask(this);
      // Convert all but the last element to avoid a trailing ","
      std::copy(args.begin(), args.end()-1,
                std::ostream_iterator<string>(oss, " "));
      oss << args.back();
      task->cmd= oss.str();
    } else return -1;
    
    assignCmd(task,args);
    {
      scoped_lock lck (*this);
      
      task->key = ++taskcnt;
      tasks.insert( std::pair<int,dmTask*>(task->key,task) );
      taskkey = task->key;
    }
    
    return taskkey;
    
    
    
  }
  
  void dmTaskExec::goCmd(int id) {
    try {
      boost::thread workerThread( taskfunc,this,id);

      // Explicit detach to work around different boost implementations of the boost::thread destructor
      workerThread.detach();
    }
    catch ( ... ) {
      dmTaskErr(this, "goCmd", "Exception when starting thread for task " << id << " on instance '" << instance << "'");
    }
    
  }
  
  
  
  /// Split che command string into the single parms
  void dmTaskExec::assignCmd(dmTask *task, std::vector<std::string> &args) {
    int i = 0;
    for(std::vector<string>::iterator it = args.begin() ; it != args.end(); ++it) {
      task->parms[i++] = strdup((*it).c_str());
    }
  }
  
  int dmTaskExec::waitResult(int taskID, int tmout) {
    
    dmTask *dt = NULL;
    
    {
      scoped_lock lck (*this);
      
      map <int, dmTask *>::iterator i = tasks.find(taskID);
      if ( i != tasks.end() ) {
        dmTaskLog(this, Logger::Lvl4, "waitResult", "Found task " << taskID);
        dt = i->second;
      }
    }
    
    if (dt) {
      dt->waitFinished(tmout);
      if (dt->finished) return 0;
      dmTaskLog(this, Logger::Lvl4, "waitResult", "Task with ID " << taskID << " has not finished in " << tmout << " seconds.");
      return 1;
    }
    else {
      dmTaskLog(this, Logger::Lvl4, "waitResult", "Task with ID " << taskID << " not found");
      return 1;
    }
    
  }
  
  int dmTaskExec::killTask(int taskID){
    
    dmTask *dt = NULL;
    {
      scoped_lock lck (*this);
      
      map <int, dmTask *>::iterator i = tasks.find(taskID);
      
      if ( i != tasks.end() ) {
        dmTaskLog(this, Logger::Lvl4, "killTask", "Found task " << taskID);
        dt = i->second;
      }
    }
    
    if (dt) {
      killTask(dt);
      return 0;
    }
    else {
      dmTaskLog(this, Logger::Lvl4, "waitTask", "Task with ID " << taskID << " not found");
      return 1;
    }
    
    return 0;
  }
  
  
  dmTask* dmTaskExec::getTask(int taskID) {
    
    scoped_lock lck (*this);
    
    map <int, dmTask *>::iterator i = tasks.find(taskID);
    if ( i != tasks.end() ) 
      return i->second;
    else 
      return NULL;
    
  }
  
  int dmTaskExec::getTaskStdout(int taskID, std::string &stdout) {
    
    scoped_lock lck (*this);
    
    map <int, dmTask *>::iterator i = tasks.find(taskID);
    if ( i != tasks.end() ) {
      boost::unique_lock <boost::mutex>  lck2 (*i->second);
      stdout = i->second->stdout;
      return 0;
    }
    else 
      return -1;
    
  }
  
  
  int dmTaskExec::killTask(dmTask *task){
    
    boost::lock_guard<dmTask> l(*task);
    
    
    if (task->finished){
      dmTaskLog(this, Logger::Lvl4, "killTask", "Task " << task->key << " already finished");
      return 0;
    } else if ( task->pid == -1) {
      dmTaskLog(this, Logger::Lvl4, "killTask", "Task " << task->key << " not yet started");
      return 0;
    } else {
      kill(task->pid, SIGKILL);
      //close fd related to the task
      for(int k=0; k<3; k++) 
        close(task->fd[k]);
      dmTaskLog(this, Logger::Lvl4, "killedTask", "Task " << task->key);
      return 0;
    }
    
  }
  
  
  
  
  void dmTaskExec::tick() {
    std::deque<dmTask *> notifq_running;
    std::deque<dmTask *> delq;
    
    int maxruntime = CFG->GetLong("glb.task.maxrunningtime", 3600);
    int purgetime = CFG->GetLong("glb.task.purgetime", 3600);
    
    dmTaskLog(this, Logger::Lvl4, "tick", "tick");
    
    {
      scoped_lock lck (*this);
      map<int, dmTask *>::iterator i, ii;
      

        for( i = tasks.begin(); i != tasks.end();  ) {
          dmTaskLog(this, Logger::Lvl4, "tick", "Found task " << i->first << " with command " << i->second->cmd);
          dmTaskLog(this, Logger::Lvl4, "tick", "The status of the task is " << i->second->finished);
          dmTaskLog(this, Logger::Lvl4, "tick", "StartTime " << i->second->starttime << " EndTime " << i->second->endtime);
          dmTaskLog(this, Logger::Lvl4, "tick", "Pid " << i->second->pid << " resultcode " << i->second->resultcode);
          time_t timenow;
          time(&timenow);
          bool terminated = false;
          
          // Treat the task object, in locked state, and accumulate events to be sent later
          {
            boost::unique_lock< boost::mutex > lck (*(i->second));
            
            if (!i->second->finished &&  (i->second->starttime< (timenow - maxruntime))) {
              dmTaskErr(this, "tick", "Task with id  " << i->first << " exceed maxrunnngtime " << (timenow - maxruntime));
              //we kill the task
              killTask(i->second);
                            
              terminated = true;
              dmTaskLog(this, Logger::Lvl3, "tick", "Task killed ");
            }
            
            //check if purgetime has exceeded and clean
            if (i->second->finished && ( i->second->endtime < (timenow - purgetime))) {
              int taskkey = i->first;
              dmTaskLog(this, Logger::Lvl4, "tick", "Task with id  " << taskkey << " to purge");
              
              //delete the task
              delq.push_back(i->second);
              dmTaskLog(this, Logger::Lvl4, "tick", "Task with id  " << taskkey << " to be deleted");
              //remove from map
              ii = i;
              ii++;
              tasks.erase(i);
              i = ii;
              
              dmTaskLog(this, Logger::Lvl3, "tick", "Task with id  " << i->first << " purged");
            } else {
              if (!terminated && !i->second->finished)
                notifq_running.push_back( i->second );
              ++i;
            }

          }
          
        }

    } // lock
    
    // Now send the accumulated notifications. We send them here in order not to keep the whole
    // object locked. Of course we are assuming that even in case of network hiccups
    // this phase ends well quicker than the purgetime
    for( std::deque<dmTask *>::iterator q = notifq_running.begin(); q != notifq_running.end(); ++q ) {
      boost::unique_lock< boost::mutex > lck (**q);
      onTaskRunning(**q);
    }
    
    int deleted = delq.size();
    {
      scoped_lock lck (*this);
      for( std::deque<dmTask *>::iterator q = delq.begin(); q != delq.end(); ++q ) {
        dmTaskLog(this, Logger::Lvl4, "tick", "Task with id  " << (*q)->key << " deleted");
        delete *q;
      }
    }
    if (deleted)
      dmTaskLog(this, Logger::Lvl4, "tick", "Deleted " << deleted << " finished tasks.");
    
  }
  
  
  
  void dmTaskExec::onTaskRunning(dmTask &task) {
    dmTaskLog(this, Logger::Lvl3, "onTaskRunning", "task " << task.key << " with command " << task.cmd);    
  }
  void dmTaskExec::onTaskCompleted(dmTask &task) {
    dmTaskLog(this, Logger::Lvl3, "onTaskCompleted", "task " << task.key << " res: " << task.resultcode << " with command " << task.cmd);
  }
  
  
}
